/*
zsfmt2.c Cu{
coded by isaku@pb4.so-net.ne.jp
*/
#include <math.h>
#include <limits.h>
#include "zsfmt.h"

#ifdef MSEQ
struct _state_t DefaultMt[MSEQ];
#else
struct _state_t DefaultMt[MSEQ2];
#endif

#define N128 ((int)(MEXP/128)+1)
#define N32 (N128*4)
#define N16 (N128*8)
#define LAG (N32>=623?11:N32>=68?7:N32>=39?5:3)
#define MID ((N32-LAG)/2)
#define MLG (MID+LAG)

#if INT_MAX <= 32767
#ifdef BIG_ENDIAN
#define BE1(X) ((X)^1)
#else
#define BE1(X) (X)
#endif
#endif

#if LONG_MAX == 0x7fffffffL
#define MSK32(X)
#define AND32(X) (X)
#else
#define MSK32(X) X&=0xffffffffUL
#define AND32(X) ((X)&0xffffffffUL)
#endif

#if MEXP == 607L
#include "z-607.h"
#elif MEXP == 1279L
#include "z-1279.h"
#elif MEXP == 2281L
#include "z-2281.h"
#elif MEXP == 4253L
#include "z-4253.h"
#elif MEXP == 11213L
#include "z-11213.h"
#elif MEXP == 19937L
#include "z-19937.h"
#elif MEXP == 44497L
#include "z-44497.h"
#elif MEXP == 86243L
#include "z-86243.h"
#elif MEXP == 132049L
#include "z-132049.h"
#elif MEXP == 216091L
#include "z-216091.h"
#else
#error MEXP ̒ls
#endif

#if MSEQ2 != 1

#ifdef HAVE_SSE2
static void(*MtFuncSSE2[])(__m128i*)={
	mfs00_,mfs01_,mfs02_,mfs03_,mfs04_,mfs05_,mfs06_,mfs07_
		,mfs08_,mfs09_,mfs10_,mfs11_,mfs12_,mfs13_,mfs14_,mfs15_
		,mfs16_,mfs17_,mfs18_,mfs19_,mfs20_,mfs21_,mfs22_,mfs23_
		,mfs24_,mfs25_,mfs26_,mfs27_,mfs28_,mfs29_,mfs30_,mfs31_
};
#endif

static void(*MtFunc[])(unsigned long*)={
	mfg00_,mfg01_,mfg02_,mfg03_,mfg04_,mfg05_,mfg06_,mfg07_
		,mfg08_,mfg09_,mfg10_,mfg11_,mfg12_,mfg13_,mfg14_,mfg15_
		,mfg16_,mfg17_,mfg18_,mfg19_,mfg20_,mfg21_,mfg22_,mfg23_
		,mfg24_,mfg25_,mfg26_,mfg27_,mfg28_,mfg29_,mfg30_,mfg31_
};

char*IdStringsTbl[]={
	ids00_,ids01_,ids02_,ids03_,ids04_,ids05_,ids06_,ids07_
		,ids08_,ids09_,ids10_,ids11_,ids12_,ids13_,ids14_,ids15_
		,ids16_,ids17_,ids18_,ids19_,ids20_,ids21_,ids22_,ids23_
		,ids24_,ids25_,ids26_,ids27_,ids28_,ids29_,ids30_,ids31_
};

#else

char*IdStringsTbl[]={ids00_};

#endif

/*
SSE2eNmWT|[g̃`FbN
*/
#ifdef HAVE_SSE2
#ifdef _MSC_VER
static int IsSSE2SFMT(void)
{
	_asm {
		push    edx
			pushfd
			pop     eax
			xor     eax, 00200000h
			push    eax
			popfd
			pushfd
			pop     edx
			cmp     eax, edx
			jnz     no_sse2
			mov     eax, 1
			cpuid
			and     edx, 04000000h
			jnz     yes_sse2
no_sse2:
		xor     eax,eax
			jmp     exit_sse2
yes_sse2:
		mov     eax,1
exit_sse2:
		pop     edx
	}
}
#else
static int IsSSE2SFMT(void) { return 1; }
#endif
#endif

/*
@̉֐
 @*/
static void period_certification(state_p mt)
{
	int i,j; long inner=0; unsigned long work,*parity;

#if MSEQ2 == 1
	parity=MtParity[0];
#else
	parity=MtParity[mt-DefaultMt];
#endif
	mt->index=N32; mt->range=0; mt->normal_sw=0;
#ifdef HAVE_SSE2
	mt->sse2=IsSSE2SFMT();
#else
	mt->sse2=0;
#endif
	for (i=0;i<4;i++) inner^=mt->u.x[i]&parity[i];
	for (i=16;i>0;i>>=1) inner^=inner>>i;
	inner&=1;
	if (inner==1) return;
	for (i=0;i<4;i++) for (j=0,work=1;j<32;j++,work<<=1)
		if (work&parity[i]) { mt->u.x[i]^=work; return; }
}

/*
@̈
 @*/
void InitMt_r(state_p mt,unsigned long s)
{
	unsigned p;

	MSK32(s);
	for (mt->u.x[0]=s,p=1;p<N32;p++)
		mt->u.x[p]=s=AND32(1812433253UL*(s^(s>>30))+p);
	period_certification(mt);
}

/*
@zgď
 @*/
void InitMtEx_r(state_p mt,unsigned long*init_key,unsigned key_len)
{
	unsigned long r,*x=mt->u.x; unsigned i,j,c;

	for (i=0;i<N32;i++) x[i]=0x8b8b8b8bUL;
	if (key_len+1>N32) c=key_len+1; else c=N32;
	r=x[0]^x[MID]^x[N32-1];
	r=AND32((r^(r>>27))*1664525UL);
	x[MID]+=r; MSK32(x[MID]); r+=key_len; MSK32(r);
	x[MLG]+=r; MSK32(x[MLG]); x[0]=r; c--;
	for (i=1,j=0;j<c&&j<key_len;j++) {
		r=x[i]^x[(i+MID)%N32]^x[(i+N32-1)%N32];
		r=AND32((r^(r>>27))*1664525UL); x[(i+MID)%N32]+=r;
		MSK32(x[(i+MID)%N32]); r+=init_key[j]+i; MSK32(r);
		x[(i+MLG)%N32]+=r; MSK32(x[(i+MLG)%N32]);
		x[i]=r; i=(i+1)%N32;
	}
	for (;j<c;j++) {
		r=x[i]^x[(i+MID)%N32]^x[(i+N32-1)%N32];
		r=AND32((r^(r>>27))*1664525UL); x[(i+MID)%N32]+=r;
		MSK32(x[(i+MID)%N32]); r+=i; MSK32(r);
		x[(i+MLG)%N32]+=r; MSK32(x[(i+MLG)%N32]);
		x[i]=r; i=(i+1)%N32;
	}
	for (j=0;j<N32;j++) {
		r=AND32(x[i]+x[(i+MID)%N32]+x[(i+N32-1)%N32]);
		r=AND32((r^(r>>27))*1566083941UL); x[(i+MID)%N32]^=r;
		r-=i; MSK32(r); x[(i+MLG)%N32]^=r; x[i]=r;
		i=(i+1)%N32;
	}
	period_certification(mt);
}

/*
@RXgN^
 @*/
#if defined(__cplusplus) && !defined(__GNUC__)
_state_t::_state_t()
{
	static volatile long seed=-1;

	if (this==DefaultMt) InitMt_r(this,1);
	else InitMt_r(this,seed--);
}
#endif

/*
@32rbgȂÅ̗{֐
 @*/
#ifdef __GNUC__
__inline__
#endif
unsigned long NextMt_r(state_p mt)
{
#if !defined(__cplusplus) || defined(__GNUC__)
	static volatile long seed=-1;

	if (mt->index==0)
	{ if (mt==DefaultMt) InitMt_r(mt,1); else InitMt_r(mt,seed--); }
#endif
	if (mt->index==N32) {
#ifdef HAVE_SSE2
#if MSEQ2 == 1
		if (mt->sse2) mfs00_(mt->u.m); else
#else
		if (mt->sse2) MtFuncSSE2[mt-DefaultMt](mt->u.m); else
#endif
#endif

#if MSEQ2 == 1
			mfg00_(mt->u.x);
#else
			MtFunc[mt-DefaultMt](mt->u.x);
#endif
		mt->index=0;
	}
	return mt->u.x[mt->index++];
}

/*
@OȏP̕_l(53bitx) 
 @*/
double NextUnif_r(state_p mt)
{
	unsigned long x=NextMt_r(mt)>>11,y=NextMt_r(mt);
	return(y*2097152.0+x)*(1.0/9007199254740992.0);
}

/*
@ۂߌ덷̂ȂOȏrange̐@@@@@@@ 
 @@@@@@@@@@@@@@@@@@@@@@@@@@@@
  @rangeȂNextInt()R{@@  @@@@@
   @for (i=0;i<10000;i++)@@@@@@@@@@ @@@@@@
	@{ x[i]=NextIntEx(100); y[i]=NextIntEx(200); }     @  
	 @Ƃ@@@@@@@@@@@@@@@@@@@@@
	  @for (i=0;i<10000;i++) x[i]=NextIntEx(100);@@@    @
	   @for (i=0;i<10000;i++) y[i]=NextIntEx(200);@@    @@
		@ƂقT{ɂȂB܂Angā@@
		 @for (i=0;i<10000;i++)@@@@@@@@@@ @@@@@@
		  @{ x[i]=NextIntEx_n(1,100); y[i]=NextIntEx_n(2,200); } 
		   @ƂĂT{ɂȂ@@@@@@@@@@@@@@@@
			@*/
long NextIntEx_r(state_p mt,long range)
{
	unsigned long y,base,remain; int shift;

	if (range<=0) return 0;
	if (range!=mt->range) {
		mt->base=mt->range=range;
		for (mt->shift=0;mt->base<=(1UL<<30);mt->shift++)
			mt->base<<=1;
	}
	for (;;) {
		y=NextMt_r(mt)>>1;
		if (y<mt->base) return(long)(y>>mt->shift);
		base=mt->base; shift=mt->shift; y-=base;
		remain=(1UL<<31)-base;
		for (;remain>=(unsigned long)range;remain-=base) {
			for (;base>remain;base>>=1) shift--;
			if (y<base) return(long)(y>>shift); else y-=base;
		}
	}
}

/*
@Rx˂̃JCQ敪z p.27 
 @*/
double NextChisq_r(state_p mt,double n)
{
	return 2*NextGamma_r(mt,0.5*n);
}

/*
@p[^̃K}z p.31 
 @*/
double NextGamma_r(state_p mt,double a)
{
	double t,u,x,y;
	if (a>1) {
		t=sqrt(2*a-1);
		do {
			do {
				do {
					x=1-NextUnif_r(mt);
					y=2*NextUnif_r(mt)-1;
				} while (x*x+y*y>1);
				y/=x; x=t*y+a-1;
			} while (x<=0);
			u=(a-1)*log(x/(a-1))-t*y;
		} while (u<-50||NextUnif_r(mt)>(1+y*y)*exp(u));
	} else {
		t=2.718281828459045235/(a+2.718281828459045235);
		do {
			if (NextUnif_r(mt)<t) {
				x=pow(NextUnif_r(mt),1/a); y=exp(-x);
			} else {
				x=1-log(1-NextUnif_r(mt)); y=pow(x,a-1);
			}
		} while (NextUnif_r(mt)>=y);
	}
	return x;
}

/*
@mo̊􉽕z p.34 
 @*/
int NextGeometric_r(state_p mt,double p)
{ return(int)ceil(log(1.0-NextUnif_r(mt))/log(1-p)); }

/*
@Opz p.89 
 @*/
double NextTriangle_r(state_p mt)
{ double a=NextUnif_r(mt),b=NextUnif_r(mt); return a-b; }

/*
@ςP̎wz p.106
 @*/
double NextExp_r(state_p mt)
{ return-log(1-NextUnif_r(mt)); }

/*
@WKz(ő6.660437) p.133
 @*/
double NextNormal_r(state_p mt)
{
	if (mt->normal_sw==0) {
		double t=sqrt(-2*log(1.0-NextUnif_r(mt)));
		double u=3.141592653589793*2*NextUnif_r(mt);
		mt->normal_save=t*sin(u); mt->normal_sw=1; return t*cos(u);
	}else{ mt->normal_sw=0; return mt->normal_save; }
}

/*
@m̃_PʃxNg p.185
 @*/
void NextUnitVect_r(state_p mt,double*v,int n)
{
	int i; double r=0;
	for (i=0;i<n;i++) { v[i]=NextNormal_r(mt); r+=v[i]*v[i]; }
	r=sqrt(r);
	for (i=0;i<n;i++) v[i]/=r;
}

/*
@p[^m,ôQz p.203 
 @*/
int NextBinomial_r(state_p mt,int n,double p)
{
	int i,r=0;
	for (i=0;i<n;i++) if (NextUnif_r(mt)<p) r++;
	return r;
}

/*
@֌Wq̂QϗʐKz p.211
 @*/
void NextBinormal_r(state_p mt,double r,double*x,double*y)
{
	double r1,r2,s;
	do {
		r1=2*NextUnif_r(mt)-1;
		r2=2*NextUnif_r(mt)-1;
		s=r1*r1+r2*r2;
	} while (s>1||s==0);
	s= -log(s)/s; r1=sqrt((1+r)*s)*r1;
	r2=sqrt((1-r)*s)*r2; *x=r1+r2; *y=r1-r2;
}

/*
@p[^`,ãx[^z p.257 
 @*/
double NextBeta_r(state_p mt,double a,double b)
{
	double temp=NextGamma_r(mt,a);
	return temp/(temp+NextGamma_r(mt,b));
}

/*
@p[^m̗ݏ敪z p.305
 @*/
double NextPower_r(state_p mt,double n)
{ return pow(NextUnif_r(mt),1.0/(n+1)); }

/*
@WXeBbNz p.313
 @*/
double NextLogistic_r(state_p mt)
{
	double r;
	do r=NextUnif_r(mt); while (r==0);
	return log(r/(1-r));
}

/*
@R[V[z p.331
 @*/
double NextCauchy_r(state_p mt)
{
	double x,y;
	do { x=1-NextUnif_r(mt); y=2*NextUnif_r(mt)-1; }
	while (x*x+y*y>1);
	return y/x;
}

/*
@Rx`,âez p.344 
 @*/
double NextFDist_r(state_p mt,double n1,double n2)
{
	double nc1=NextChisq_r(mt,n1),nc2=NextChisq_r(mt,n2);
	return (nc1*n2)/(nc2*n1);
}

/*
@σɂ̃|A\z p.412
 @*/
int NextPoisson_r(state_p mt,double lambda)
{
	int k; lambda=exp(lambda)*NextUnif_r(mt);
	for (k=0;lambda>1;k++) lambda*=NextUnif_r(mt);
	return k;
}

/*
@Rxm̂z p.428
 @*/
double NextTDist_r(state_p mt,double n)
{
	double a,b,c;
	if (n<=2) {
		do a=NextChisq_r(mt,n); while (a==0);
		return NextNormal_r(mt)/sqrt(a/n);
	}
	do {
		a=NextNormal_r(mt); b=a*a/(n-2);
		c=log(1-NextUnif_r(mt))/(1-0.5*n);
	} while (exp(-b-c)>1-b);
	return a/sqrt((1-2.0/n)*(1-b));
}

/*
@p[^̃Cuz p.431
 @*/
double NextWeibull_r(state_p mt,double alpha)
{ return pow(-log(1-NextUnif_r(mt)),1/alpha); }
